import flet as ft

from django.utils.translation import gettext as _

FIELDS_MODELS = {
    'IntegerField': 'text',
    'DurationField': None,
    'ManyToOneRel': None,
    'DateTimeField': 'date',
    'FileField': 'file',
    'TextField': 'text',
    'ImageField': None,
    'ForeignKey': None,
    'CharField': 'text',
    'BigIntegerField': 'text',
}


class ModelFormControl(ft.UserControl):

    def __init__(
            self,
            obj=None,
            title=None,
            **form_params
    ):
        super().__init__()

        self.obj = obj
        self.title = title or ft.Text(str(obj))
        self.content = ft.Text("PLACE FOR YOUR FORM")

    def build(self):
        return ft.Card(
            ft.Container(
                ft.Column([
                    ft.Text(self.title, style=ft.TextThemeStyle.HEADLINE_SMALL),
                    self.content,
                ]),
                padding=10,
            ),
        )

    def refresh_data(self, *args):
        self.update()

    def did_mount(self):
        self.refresh_data()



from dataclasses import dataclass, field
from abc import abstractmethod
from collections import defaultdict
from importlib import import_module
from typing import Type, Iterable, TypeVar, Callable, Dict, Optional, List
from typing import Union

import flet as ft
from flet_core import Control

from django.conf import settings

from .middlewares import GenericMiddleware, urls_middleware, simple_view_middleware
from .navigation import Destiny, Navigare
from .views import GenericViewFactory
from .types import CLIENT_CLASS


@dataclass
class GenericApp:
    middlewares: list[Type[GenericMiddleware]] = field(default_factory=list)
    urls: Optional[Iterable] = None
    controls: Optional[list[Control]] = None
    view: Optional[Callable[[CLIENT_CLASS], ft.View]] = None
    text: Optional[str] = None
    destinations: Optional[List[Destiny]] = None
    client_class: Optional[Type[CLIENT_CLASS]] = None
    view_params: dict = field(default_factory=dict)
    init_route: str = '/'
    view_factory: Callable[
        [CLIENT_CLASS],
        Callable[
            [list[Control], ...],
            ft.View
        ],
    ] = GenericViewFactory

    def __call__(self, page):
        page.scroll = ft.ScrollMode.ALWAYS
        return self.client_class(app=self, page=page)

    def __post_init__(self):

        try:
            engine = import_module(settings.SESSION_ENGINE)
            # pylint: disable=E0116
            # value is a class
            self.SessionStore = engine.SessionStore
        except Exception as _:
            self.SessionStore = lambda: {}

        if self.client_class is None:
            self.client_class = GenericClient

        if self.urls is not None:
            self.middlewares.append(urls_middleware(urls=tuple(self.urls)))

        simple_views_params = [self.controls, self.view, self.text]
        if any(simple_views_params):
            self.middlewares.append(simple_view_middleware(controls=self.controls, view=self.view, text=self.text))


@dataclass
class GenericClient:
    app: GenericApp
    page: ft.Page
    navigation: Optional[Navigare] = None
    on_route_change: Optional[Callable[[ft.RouteChangeEvent], None]] = None
    on_view_pop: Optional[Callable[[ft.ViewPopEvent], None]] = None
    __current_route: Optional[str] = None
    __bs: Optional[ft.BottomSheet] = None
    __dialog: Optional[ft.AlertDialog] = None

    def __post_init__(self) -> None:

        if self.app.destinations:
            self.navigation = self.navigation or Navigare(self, self.app.destinations)

        self.session = self.app.SessionStore()
        self.page.on_route_change = self._route_change
        self.page.on_view_pop = self._pop_view

        self.page.views.pop()  # remove empty view from top

        self.route = self.app.init_route

    def _route_change(self, e: ft.RouteChangeEvent) -> None:
        if self.on_route_change is None:
            self.route = e.route
        else:
            self.on_route_change(e)

    def _pop_view(self, e: ft.ViewPopEvent) -> None:
        if self.on_view_pop is None:
            self.pop()
        else:
            self.on_view_pop(e)

    @property
    def bs(self):
        def on_dismiss(e):
            self.page.overlay.pop(self.__bs)
            self.__bs = None

        if self.__bs is None:
            self.__bs = ft.BottomSheet(
                open=False,
                on_dismiss=on_dismiss,
            )
            self.page.overlay.append(self.__bs)
            self.__bs.show = self.bs_show
            self.__bs.hide = self.bs_hide
        return self.__bs

    @bs.setter
    def bs(self, content: ft.Control):
        self.bs.content = content
        self.bs.open = True
        self.page.update()

    def bs_show(self):
        self.bs.open = True
        self.page.update()

    def bs_hide(self):
        self.bs.open = False
        self.page.update()

    @property
    def dialog(self):
        def on_dismiss(e):
            self.page.dialog = None
            self.page.update()
            self.__dialog = None

        if self.__dialog is None:
            self.__dialog = ft.AlertDialog(
                modal=True,
                actions_alignment=ft.MainAxisAlignment.END,
                on_dismiss=on_dismiss,
            )
            self.page.dialog = self.__dialog
            self.__dialog.show = self.dialog_show
            self.__dialog.hide = self.dialog_hide
        return self.__dialog

    @dialog.setter
    def dialog(self, new_dialog: ft.AlertDialog):
        self.__dialog = new_dialog
        self.page.dialog = self.__dialog
        self.page.update()

    def update_dialog(self, content: ft.Control, actions=None, show=None, title=None):
        self.dialog.content = content
        self.dialog.actions = actions or []

        if title is not None:
            self.dialog.title = ft.Text(title)

        if show is not None:
            self.dialog.open = show

        self.page.update()

    def dialog_show(self):
        self.dialog.open = True
        self.page.update()

    def dialog_hide(self):
        self.dialog.open = False
        self.page.update()

    @property
    def get_view(self):
        return self.app.view_factory(self, self.app.view_params)

    @property
    def route(self):
        return self.__current_route

    def parse_route(self, route: str) -> Callable:
        __middlewares = []
        if route[0] != '/':  # absolute paths only
            route = f"/{route}"
        view = None
        for middleware_class in self.app.middlewares:
            middleware = middleware_class(page=self, route=route)
            __middlewares.append(middleware)
            view = middleware.parse_route()
            if view is not None:
                break
        while len(__middlewares):
            middleware = __middlewares.pop()
            view = middleware.parse_view(view)

        return view

    @route.setter
    def route(self, route: str):
        self.__current_route = route
        view = self.parse_route(route)
        self.append_view(view)
        self.update()

    def append_view(self, view: Callable):
        view_control = view(self)
        self.page.views.append(view_control)

    def update(self, *controls):
        self.page.update(*controls)

    def go(self, route: str):
        self.route = route

    def pop(self) -> None:
        if len(self.page.views) == 1:
            return
        self.page.views.pop()
        top_view = self.page.views[-1]
        self.page.go(top_view.route)



from typing import TypeVar

DESTINATION_CLASS = TypeVar('DESTINATION_CLASS', bound='Destiny')
CLIENT_CLASS = TypeVar('CLIENT_CLASS', bound='GenericClient')
VIEW_FACTORY = TypeVar('VIEW_FACTORY', bound='GenericViewFactory')



from abc import ABC
from dataclasses import dataclass, field
from typing import Optional, List, Callable

import flet as ft
from flet_core import Control

from flet_django.types import CLIENT_CLASS, DESTINATION_CLASS


@dataclass
class Navigare(ABC):
    page: CLIENT_CLASS
    destinations: Optional[List[DESTINATION_CLASS]] = field(default_factory=list)
    selected_index: Optional[int] = None

    def on_change(self) -> Callable:
        def __on_change(e):
            """
            Define the NavigationBar control behaviour.

            :param e: event
            :return: return routing url depending on
                a chosen button of NavigationBar control
            """
            try:
                index = int(e.data)
                self.selected_index = index
                destiny = self.destinations[index]
                self.page.route = destiny.route

            except ValueError:
                print(f"Can't parse index value:{e.data}")
            except IndexError:
                print(f"Can't found destiny for index:{e.data}")

        return __on_change

    def get_bar(self, *args, **kwagrs):
        destinations = []
        index = -1
        for destiny in self.destinations:
            if destiny.nav_bar:
                destinations.append(destiny.get_bar())

                # for a case if selected index is not set on beginning
                # recognize current view and set it as selected
                index += 1
                if self.selected_index is None and destiny.route == self.page.route:
                    self.selected_index = index

        new_kwargs = dict(
            on_change=self.on_change(),
            destinations=destinations,
            selected_index=self.selected_index
        )
        new_kwargs.update(kwagrs)
        return ft.NavigationBar(*args, **new_kwargs)

    def get_rail(self, *args, **kwagrs):
        new_kwargs = dict(
            on_change=self.on_change(),
            destinations=[destiny.get_rail() for destiny in self.destinations if destiny.nav_rail]
        )
        new_kwargs.update(kwagrs)
        return ft.NavigationRail(*args, **new_kwargs)

    def get_actions(self, button_factory: Callable = None, **kwargs):
        return [destiny.get_button(self.page, button_factory, **kwargs) for destiny in self.destinations if destiny.action]


@dataclass
class Destiny(ABC):
    route: str
    icon: Optional[str] = None
    icon_content: Optional[Control] = None
    selected_icon: Optional[str] = None
    selected_icon_content: Optional[Control] = None
    label: Optional[str] = None
    label_content: Optional[Control] = None,
    padding: ft.PaddingValue = None,
    nav_bar: bool = True
    nav_rail: bool = True
    action: bool = True

    def get_bar(self, *args, **kwagrs):
        new_kwargs = dict(
            icon=self.icon,
            icon_content=self.icon_content,
            selected_icon=self.selected_icon,
            selected_icon_content=self.selected_icon_content,
            label=self.label
        )
        new_kwargs.update(kwagrs)
        return ft.NavigationDestination(*args, **new_kwargs)

    def get_rail(self, *args, **kwagrs):
        new_kwargs = dict(
            icon=self.icon,
            icon_content=self.icon_content,
            selected_icon=self.selected_icon,
            selected_icon_content=self.selected_icon_content,
            label=self.label,
            label_content=self.label_content,
            padding=self.padding
        )
        new_kwargs.update(kwagrs)
        return ft.NavigationRailDestination(*args, **new_kwargs)

    def get_button(self, page, button_factory: Optional[Callable] = None, **kwargs):

        def __button_factory(destiny, **button_kwargs):
            selected: bool = destiny.route == page.route

            def __on_click(e):
                if selected:
                    page.pop()
                else:
                    page.route = destiny.route

            new_kwargs = dict(
                icon=destiny.icon,
                selected_icon=destiny.selected_icon,
                tooltip=destiny.label,
                selected=selected,
                on_click=__on_click
            )
            new_kwargs.update(button_kwargs)
            return ft.IconButton(**new_kwargs)

        button_factory = button_factory or __button_factory

        return button_factory(self, **kwargs)



from abc import ABC
from dataclasses import dataclass, field
from typing import Callable, Iterable, Dict

import flet as ft

@dataclass
class FtViewRoute(ABC):
    route: str
    name: str
    control: Callable
    args: Iterable = field(default_factory=list)
    kwargs: Dict = field(default_factory=dict)

    def __call__(self, page, *args, **kwargs):
        args += self.args
        kwargs.update(self.kwargs)
        result = self.control(page, *args, **kwargs)
        if isinstance(result, ft.View):
            result.route = self.route
        return result



from dataclasses import dataclass
from dataclasses import field
from typing import Optional, Type, Callable
from typing import Union

import flet as ft
from flet_core import Control

from flet_django.types import CLIENT_CLASS


def get_app_bar(page, title: str = '', action_params: Optional[dict] = None, **kwargs) -> ft.AppBar:
    action_params = action_params or {}

    new_kwargs = dict(
        title=ft.Text(value=title),
        actions=page.navigation.get_actions(**action_params)
    )
    new_kwargs.update(kwargs)

    return ft.AppBar(**new_kwargs)


@dataclass
class GenericViewFactory:
    client: CLIENT_CLASS
    kwargs: dict = field(default_factory=dict)

    def get_view(self, controls, **kwargs):
        return ft.View(controls=controls, **kwargs)

    def set_nav_bar(self, controls, **nav_bar_params):
        if self.client.navigation:
            nav_bar = self.nav_bar_factory(**nav_bar_params)
            controls.append(nav_bar)
        return controls

    def nav_bar_factory(self, **nav_bar_params) -> ft.NavigationBar:
        return self.client.navigation.get_bar(**nav_bar_params)

    def set_app_bar(self, controls, **app_bar_params):
        if app_bar_params:
            app_bar = self.app_bar_factory(**app_bar_params)
            controls = [app_bar, ] + controls
        return controls

    def app_bar_factory(self, **app_bar_params) -> ft.AppBar:
        return get_app_bar(self.client, **app_bar_params)

    def __call__(self,
                 controls: list[Control],
                 nav_bar_params: Optional[dict] = None,
                 app_bar_params: Optional[dict] = None,
                 **kwargs,
                 ):

        new_kwargs = dict(
            vertical_alignment=ft.MainAxisAlignment.CENTER,
            horizontal_alignment=ft.CrossAxisAlignment.CENTER
        )
        new_kwargs.update(self.kwargs)
        new_kwargs.update(kwargs)

        nav_bar_params = nav_bar_params or {}
        app_bar_params = app_bar_params or {}
        if 'nav_bar_params' in new_kwargs:
            current_nav_bar_params = nav_bar_params
            nav_bar_params = new_kwargs.pop('nav_bar_params')
            nav_bar_params.update(current_nav_bar_params)

        if 'app_bar_params' in self.kwargs:
            current_app_bar_params = app_bar_params
            app_bar_params = new_kwargs.pop('app_bar_params')
            app_bar_params.update(current_app_bar_params)

        view = self.get_view(controls=controls, **new_kwargs)
        view.controls = self.set_nav_bar(view.controls, **nav_bar_params)
        view.controls = self.set_app_bar(view.controls, **app_bar_params)

        return view


#################################################
#                   OLD CODE                    #
#################################################


def ft_view(
    page: CLIENT_CLASS,
    controls: list[Control],
    nav_bar_params: Optional[dict] = None,
    app_bar_params: Optional[dict] = None,
    app_bar_factory: Callable = get_app_bar,
    **kwargs,
):
    new_kwargs = dict(
        vertical_alignment=ft.MainAxisAlignment.CENTER,
        horizontal_alignment=ft.CrossAxisAlignment.CENTER
    )

    if page.app.view_params:
        new_kwargs.update(page.app.view_params)

    if 'nav_bar_params' in new_kwargs:
        current_nav_bar_params = nav_bar_params or {}
        nav_bar_params = new_kwargs.pop('nav_bar_params')
        nav_bar_params.update(current_nav_bar_params)

    if 'app_bar_params' in new_kwargs:
        current_app_bar_params = app_bar_params or {}
        app_bar_params = new_kwargs.pop('app_bar_params')
        app_bar_params.update(current_app_bar_params)

    new_kwargs.update(kwargs)

    if page.navigation:
        nav_bar_params = nav_bar_params or {}
        controls.append(page.navigation.get_bar(**nav_bar_params))

    if app_bar_params:
        app_bar = app_bar_factory(page, **app_bar_params)
        controls = [app_bar, ] + controls

    return ft.View(controls=controls, **new_kwargs)



from .middlewares import simple_view_middleware, urls_middleware, GenericMiddleware
from .navigation import Destiny, Navigare
from .pages import GenericApp, GenericClient
from .views import ft_view, GenericViewFactory



from abc import ABC
from dataclasses import dataclass
from typing import Callable
from typing import Optional

from django.urls import resolve, Resolver404

import flet as ft
from flet_core import Control

from .routes import FtViewRoute
from .types import CLIENT_CLASS


@dataclass
class GenericMiddleware(ABC):
    page: CLIENT_CLASS
    route: str

    def parse_route(self):
        return None

    def parse_view(self, view):
        return view


def simple_view_middleware(
    controls: Optional[list[Control]] = None,
    text: Optional[str] = None,
    view: Optional[Callable[[CLIENT_CLASS], ft.View]] = None,
    **kwargs
):
    if controls is None:
        controls = [ft.Text(text)]
    if view is None:
        def __view(page: CLIENT_CLASS):
            return page.get_view(
                controls=controls,
                **kwargs
            )

        view = __view

    class SimpeViewMiddleware(GenericMiddleware):

        def parse_route(self):
            return view

    return SimpeViewMiddleware


def urls_middleware(urls: tuple):
    class UrlsMiddleware(GenericMiddleware):
        def parse_route(self):
            try:
                result = resolve(self.route, urls)
                return FtViewRoute(
                    control=result.func,
                    name=result.url_name,
                    route=self.route,
                    args=result.args,
                    kwargs=result.kwargs
                )
            except Resolver404:
                return None

    return UrlsMiddleware



import flet as ft
from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from main_app import main


class Command(BaseCommand):
    help = _('Run flutter app.')

    def add_arguments(self, parser):
        parser.add_argument(
            '--view',
            type=str,
            default=ft.FLET_APP,
            required=False,
            help=_("Chose view from web_browser, flet_app and flet_app_hidden. Default is flet_app.")
        )
        parser.add_argument(
            '--port',
            type=str,
            default=8085,
            required=False,
            help=_("Chose port. Default is 8085.")
        )
        parser.add_argument(
            '--host',
            type=str,
            default="127.0.0.1",
            required=False,
            help=_("Chose host. Default is 127.0.0.1.")
        )

    def handle(self, *args, **options):

        view = options["view"]

        self.stdout.write(self.style.MIGRATE_HEADING('Let run flutter app.'))
        ft.app(
            target=main,
            port=options["port"], 
            view=view,
            host=options["host"],
            assets_dir="assets"
        )
        self.stdout.write(self.style.SUCCESS('Finish running flutter app.'))



import flet as ft

from django.utils.translation import gettext as _
from django.core.paginator import Paginator
from django.db.models import Q

ERROR_MSG = "-Err-"

FIELDS_MODELS = {
    'IntegerField': 'text',
    'DurationField': None,
    'ManyToOneRel': None,
    'DateTimeField': 'date',
    'FileField': 'file',
    'TextField': 'text',
    'ImageField': None,
    'ForeignKey': None,
    'CharField': 'text',
    'BigIntegerField': 'text',
    'BooleanField': 'bool',
}


# Data column
class Col:

    def __init__(self, name: str, label, field_type=None, width=None, on_sort=None):
        if isinstance(label, ft.Control):
            self.label = label
        else:
            self.label = ft.Text(label)
        self.name = name
        self.width = width
        self.type = field_type
        self.on_sort = on_sort

    def build_header(self):
        return ft.DataColumn(
            label=self.label,
            on_sort=self.on_sort,
        )

    def build_cell(self, obj):
        return ft.DataCell(
            content=ft.Text(
                str(
                    getattr(
                        obj,
                        self.name,
                        _("cell_error")
                    )
                ),
                width=self.width
            )
        )


def text_filter_kwargs(field_name: str, value: str):
    return {
        f"{field_name}__icontains": value.lower()
    }


class ModelTableControl(ft.UserControl):

    def __init__(
        self,
        model,
        columns=None,
        order=None,
        rows_per_page=10,
        title=None,
        filters=None,
        auto_filter=True,
        get_data_table=None,
        **data_table_params
    ):
        super().__init__()

        self.model = model
        self.order = order
        self.current_page = 1
        self.rows_per_page = rows_per_page
        self.num_pages = 1
        self.num_rows = 0
        self.title = title or model.__name__
        self.filters = filters
        self.auto_filter = auto_filter
        self.data_table_params = data_table_params

        self.v_current_page = ft.Text(str(self.current_page))
        self.v_search = ft.TextField(
            hint_text=_("Search..."),
            dense=True,
            on_change=self.on_search_auto,
            prefix_icon=ft.icons.SEARCH,
            border=ft.InputBorder.NONE,
            filled=False,
            on_submit=self.on_search_submit
        )
        self.v_count = ft.Text()

        if columns is None:
            self.columns = list(self.get_columns())
        else:
            self.columns = columns

        self.fields_indexes = {col.name: i for i, col in enumerate(self.columns)}

        if get_data_table is None:
            self.data_table = ft.DataTable(
                columns=[col.build_header() for col in self.columns],
                rows=[],
                **data_table_params
            )
        else:
            self.data_table = get_data_table(self)

    def get_columns(self):

        fields = self.model._meta.get_fields()

        def __on_sort(e):
            sorted_field_name = self.columns[e.column_index].name
            self.order = sorted_field_name if e.ascending else f"-{sorted_field_name}"
            self.refresh_data()

        for field in fields:
            class_name = field.__class__.__name__
            field_type = FIELDS_MODELS.get(class_name, None)
            if field_type is not None:
                name = field.name
                label = field.verbose_name or field.name

                yield Col(name=name, label=label, field_type=field_type, on_sort=__on_sort)

    def set_page(self, page=None, delta=0):
        if page is not None:
            self.current_page = page
        elif delta:
            self.current_page += delta
        else:
            return
        self.refresh_data()

    def next_page(self, e):
        self.set_page(delta=1)

    def prev_page(self, e):
        if self.current_page > 1:
            self.set_page(delta=-1)

    def goto_first_page(self, e):
        self.set_page(page=1)

    def goto_last_page(self, e):
        self.set_page(page=self.num_pages)

    def build_rows(self):
        qs = self.model.objects.all()

        if self.order:
            qs = qs.order_by(self.order)

        if self.v_search.value:
            q = Q()
            for column in self.columns:
                if column.type == "text":
                    q |= Q(**text_filter_kwargs(column.name, self.v_search.value))
            qs = qs.filter(q)

        if self.filters is not None:
            qs = qs.filter(**self.filters)

        self.num_rows = qs.count()
        paginator = Paginator(qs, self.rows_per_page)
        p_int, p_add = divmod(self.num_rows, self.rows_per_page)
        self.num_pages = p_int + (1 if p_add else 0)
        page_qs = paginator.page(self.current_page)
        # Load
        return [ft.DataRow(cells=[col.build_cell(obj) for col in self.columns])
                for obj in page_qs]

    def build(self):
        return ft.Card(
            ft.Container(
                ft.Column(
                    [
                        ft.Text(self.title, style=ft.TextThemeStyle.HEADLINE_SMALL),
                        ft.Row(
                            controls=[self.data_table],
                            scroll=ft.ScrollMode.ALWAYS,
                            spacing=10,
                        ),
                        ft.Row([
                            ft.IconButton(ft.icons.KEYBOARD_DOUBLE_ARROW_LEFT, on_click=self.goto_first_page,
                                          tooltip=_("First page")),
                            ft.IconButton(ft.icons.KEYBOARD_ARROW_LEFT, on_click=self.prev_page,
                                          tooltip=_("Prev page")),
                            self.v_current_page,
                            ft.IconButton(ft.icons.KEYBOARD_ARROW_RIGHT, on_click=self.next_page,
                                          tooltip=_("Next page")),
                            ft.IconButton(ft.icons.KEYBOARD_DOUBLE_ARROW_RIGHT, on_click=self.goto_last_page,
                                          tooltip=_("Last page")),
                            self.v_search,
                            self.v_count,
                            ft.IconButton(ft.icons.REFRESH, on_click=self.refresh_data, tooltip=_("Refresh")),
                            ft.Slider(
                                min=10,
                                max=90,
                                divisions=8,
                                value=self.rows_per_page,
                                label=_("{value} rows per page"),
                                on_change=self.on_per_page_changed
                            ),
                        ]),
                    ],
                    scroll=ft.ScrollMode.ALWAYS
                ),
                padding=25,
            ),
        )

    def on_search_auto(self, e):
        if self.auto_filter:
            self.goto_first_page(e)

    def on_search_submit(self, e):
        self.goto_first_page(e)

    def on_per_page_changed(self, e):
        self.rows_per_page = int(e.control.value)
        self.refresh_data()

    def refresh_data(self, *args):
        self.data_table.rows = self.build_rows()
        if self.order:
            ascending = self.order[0] != '-'
            field_name = self.order if ascending else self.order[1:]
            order_index = self.fields_indexes[field_name]
            self.data_table.sort_ascending = ascending
            self.data_table.sort_column_index = order_index
        else:
            self.data_table.sort_ascending = None
            self.data_table.sort_column_index = None
        self.v_count.value = _("{} rows").format(self.num_rows)
        self.v_current_page.value = f"{self.current_page}/{self.num_pages}"
        self.update()

    def did_mount(self):
        self.refresh_data()
